"""
@AUTHOR: nuthanmunaiah
"""

import nltk


class PosTagger(object):
    """
    Given a list of tokens, return a list of tuples of the form:
    (token, part-of-speech-tag)
    """
    def __init__(self, tokens):
        """ Constructor. """
        self.tokens = tokens

    def execute(self):
        """
        Given a list of tokens, return a list of tuples of the form:
        (token, part-of-speech-tag)
        """
        return nltk.pos_tag(self.tokens)



from nltk.chunk import ChunkParserI, tree2conlltags as to_tags
from nltk.corpus import treebank_chunk, conll2000
from nltk.tag import UnigramTagger, BigramTagger


def tag_chunks(chunk_sents):
    tag_sents = [to_tags(tree) for tree in chunk_sents]
    return [[(t, c) for (w, t, c) in chunk_tags] for chunk_tags in tag_sents]


CHUNKS = tag_chunks(treebank_chunk.chunked_sents()) + \
         tag_chunks(conll2000.chunked_sents())
TAGGER = BigramTagger(CHUNKS, backoff=UnigramTagger(CHUNKS))


class ChunkTagger(ChunkParserI):
    def parse(self, tokens):
        (tokens, tags) = zip(*tokens)
        chunks = TAGGER.tag(tags)
        return [(token, chunk[1]) for (token, chunk) in zip(tokens, chunks)]



from . import chunktagger, lemmatizer, postagger, stemmer, tokenizer


class Summarizer(object):
    def __init__(self, text):
        self.text = text

    def execute(self):
        tokens = tokenizer.NLTKTokenizer(self.text).execute()
        stems = stemmer.Stemmer(tokens).execute()
        pos = postagger.PosTagger(tokens).execute()
        chunk = chunktagger.ChunkTagger().parse(pos)

        summary = zip(tokens, stems, pos, chunk)
        return [(t, s, p[1], c[1]) for (t, s, p, c) in summary]



import nltk


class Stemmer(object):
    """
    Given a list of tokens, return a list of stems associated with those tokens
    """
    def __init__(self, tokens):
        """ Constructor. """
        self.stemmer = nltk.PorterStemmer()
        self.tokens = tokens

    def execute(self):
        """
        Given a list of tokens, return a list of stems associated with those
        tokens.
        """
        return [self.stemmer.stem(token).lower() for token in self.tokens]



from nltk.tokenize import word_tokenize


class Tokenizer(object):
    """ Interface. """
    def __init__(self, text):
        """ Constructor. """
        self.text = text

    def execute(self):
        """ Raises NotImplementedError. """
        raise NotImplementedError("Tokenizer is an abstract class. In must be "
                                  "implemented by another class. Try using "
                                  "the NLTKTokenizer.")


class NLTKTokenizer(Tokenizer):
    """ Implements Tokenizer. """
    def __init__(self, text):
        """ Constructor. """
        super(NLTKTokenizer, self).__init__(text)

    def execute(self):
        """ Return a list of all tokens within the specified string. """
        return word_tokenize(self.text)



import csv
import re

from . import helpers

__all__ = ['Word', 'Words']


class Word(object):
    @classmethod
    def from_line(cls, line):
        instance = cls()

        components = next(csv.reader([line], delimiter='\t'))

        instance.word = components[1]
        instance.root = components[2]
        instance.pos = components[3]
        instance.chunk = None
        instance.label = components[4]
        instance.label = re.sub(r"^(B-|I-)", "", instance.label).upper()
        instance.multi_label = components[5]
        instance.binary_label = 'C' if instance.label in ['O', 'C'] else 'U'
        instance.features = dict()
        for feature in components[6:]:
            feature = re.sub(r"\|\|", "|", feature)
            index = feature.rindex(':')
            instance.features[feature[:index]] = float(feature[index + 1:])
        instance.context = None

        return instance

    @classmethod
    def from_group(cls, group):
        instance = cls()

        word, root, pos, chunk = group

        instance.word = word
        instance.root = root
        instance.pos = pos
        instance.chunk = chunk
        instance.label = None
        instance.multi_label = None
        instance.binary_label = None
        instance.features = None
        instance.context = None

        return instance

    def get_features(self):
        return self.features if self.features else self._get_features()

    def get_label(self, binary=True):
        return self.binary_label if binary else self.multi_label

    def _get_features(self):
        features = list()

        features.extend(self.get_typeonefeatures())
        features.extend(self.get_typetwofeatures())
        features.extend(self.get_typethreefeatures())
        features.extend(self.get_typefourfeatures())
        features.extend(self.get_typefivefeatures())
        features.extend(self.get_typesixfeatures())
        features.extend(self.get_typesevenfeatures())

        return dict(zip(features, [1.0] * len(features)))

    def get_typeonefeatures(self):
        '''Return prefixes of lengths 3, 4 and 5

        Returns
        ------
        features : list
            A list of prefixes of lengths 3, 4 and 5 of the current word.

        Example
        -------

        If 'Distinct' is the current word, then ['prefix_3_Dis',
        'prefix_4_Dist', 'prefix_5_Disti'] is returned.
        '''
        return ['prefix_{}_{}'.format(i, self.word[:i]) for i in [3, 4, 5]]

    def get_typetwofeatures(self):
        '''Return suffixes of lengths 3, 4 and 5

        Returns
        ------
        features : list
            A list of suffixes of lengths 3, 4 and 5 of the current word.

        Example
        -------

        If 'Distinct' is the current word, then ['suffix_3_nct',
        'suffix_4_inct', 'suffix_5_tinct'] is returned.
        '''
        return ['suffix_{}_{}'.format(i, self.word[-i:]) for i in [3, 4, 5]]

    def get_typethreefeatures(self):
        '''Return root (stem or lemma) of words in a window length of 2

        Returns
        -------
        features : list
            A list containing the root of the current word and that of two
            words before and after the current word.

        Example
        -------

        If 'Regulating' is the current word in the context 'Cells in Regulating
        Cellular Immunity', then ['root_-2_cell', 'root_-1_in',
        'root_0_regulate, 'root_1_cellular', 'root_2_immun'] is returned.
        '''
        features = ['root_0_{}'.format(self.root)]
        features += [
                'root_{}_{}'.format(offset, word.root)
                for (offset, word) in self.context.items()
                if word is not None
            ]
        return features

    def get_typefourfeatures(self):
        '''Return pattern prefixes of words in a window length of 1

        Returns
        -------
        features : list
            A list containing pattern prefix of the current word and that of
            one word before and after the current word. In addition to the
            pattern prefixes, the first character of the pattern prefix of the
            current word is also included in the list.

        Example
        -------

        If 'Regulating' is the current word in the context 'Cells in Regulating
        Cellular Immunity', the ['pattern_-1_a', 'pattern_0_Aa',
        'pattern_1_Aa', 'pattern_prefix_A'] is returned.
        '''
        features = [
                'pattern_0_{}'.format(self.pattern),
                'pattern_prefix_{}'.format(self.pattern[0])
            ]
        # TODO: How to avoid hard coding [-1, 1]?
        features += [
                'pattern_{}_{}'.format(offset, word.pattern)
                for (offset, word) in self.context.items()
                if offset in [-1, 1] and word is not None
            ]
        return features

    def get_typefivefeatures(self):
        '''Return part of speech tag of words in a window length of 2

        Returns
        -------
        features : list
            A list containing the part of speech tag of the current word and
            that of two words before and after the current word.

        Example
        -------

        If 'Regulating' is the current word in the context 'Cells (NNS) in (IN)
        Regulating (VBG) Cellular (JJ) Immunity (NN)', then ['pos_-2_NNS',
        'pos_-1_IN', 'pos_0_VBG', 'pos_1_JJ', 'pos_2_NN']
        '''
        features = ['pos_0_{}'.format(self.pos)]
        features += [
                'pos_{}_{}'.format(offset, word.pos)
                for (offset, word) in self.context.items()
                if word is not None
            ]
        return features

    def get_typesixfeatures(self):
        '''Return chunk tag of words in a window length of 2

        Returns
        -------
        features : list
            A list containing the chunk tag of the current word and that of two
            words before and after the current word.

        Example
        -------

        If 'Regulating' is the current word in the context 'Cells (I-np) in
        (B-pp) Regulating (B-vp) Cellular (B-np) Immunity (I-np)', then
        ['chunk_-2_I-np', 'chunk_-1_B-pp', 'chunk_0_B-vp', 'chunk_1_B-np',
        'chunk_2_I-np'] is returned.
        '''
        features = ['chunk_0_{}'.format(self.chunk)]
        features += [
                'chunk_{}_{}'.format(offset, word.chunk)
                for (offset, word) in self.context.items()
                if word is not None
            ]
        return features

    def get_typesevenfeatures(self):
        '''Return combinations of root, chunk and part of speech tag

        Returns
        -------
        features : list
            A list containing (1) combination of root and chunk tag and part of
            speech tag of current word, (2) combination of chunk tag of current
            word and root of one word before and after the current word, and
            (3) combination of part of speech tag of current word and root of
            one word before and after the current word.

        Example
        -------

        If 'Regulating' is the current word in the context 'Cells (NNS|I-np) in
        (IN|B-pp) Regulating (VBG|B-vp) Cellular (JJ|B-np) Immunity (NN|I-np)',
        then ['R_0_regulate_C_0_B-vp', 'R_0_regulate_P_0_VBG',
        'R_-1_in_C_0_B-vp', 'R_1_cellular_C_0_B_vp', 'R_-1_in_P_0_VBG',
        'R_1_cellular_P_0_VBG']
        '''
        features = [
                'R_0_{}_C_0_{}'.format(self.root, self.chunk),
                'R_0_{}_P_0_{}'.format(self.root, self.pos)
            ]
        # TODO: How to avoid hard coding [-1, 1]?
        features += [
                'R_{}_{}_C_0_{}'.format(offset, word.root, self.chunk)
                for (offset, word) in self.context.items()
                if offset in [-1, 1] and word is not None
            ]
        features += [
                'R_{}_{}_P_0_{}'.format(offset, word.root, self.pos)
                for (offset, word) in self.context.items()
                if offset in [-1, 1] and word is not None
            ]
        return features

    @property
    def pattern(self):
        return helpers.get_wordpattern(self.word)

    def __repr__(self):
        return self.word

    def __str__(self):
        return self.word


class Words(object):
    @classmethod
    def from_lines(cls, lines):
        instance = cls()

        instance.words = list()
        for line in lines:
            if line.strip() == '':
                continue
            instance.words.append(Word.from_line(line))

        return instance

    @classmethod
    def from_groups(cls, groups):
        instance = cls()

        instance.words = list()
        for group in groups:
            instance.words.append(Word.from_group(group))
        instance.establish_context()

        return instance

    def establish_context(self):
        for (index, word) in enumerate(self.words):
            word.context = helpers.get_context(index, self.words)

    def get_words(self):
        return self.words

    def get_data(self, binary=True):
        X, y, z = list(), list(), list()
        for word in self.words:
            X.append(word.get_features())
            y.append(word.get_label(binary))
            z.append(word.word)
        return X, y, z

    def get_features(self):
        features = []
        for word in self.words:
            features.append(word.get_features())
        return features

    def get_labels(self, binary=True):
        labels = []
        for word in self.words:
            labels.append(word.get_label(binary=binary))

        return labels



from pkg_resources import resource_filename

BCLASS_CLASSIFIER_PATH = resource_filename('uncertainty', 'models/bclass.p')
MCLASS_CLASSIFIER_PATH = resource_filename('uncertainty', 'models/mclass.p')
VECTORIZER_PATH = resource_filename('uncertainty', 'vectorizers/vectorizer.p')



from urllib.request import urlretrieve
from urllib.parse import urljoin

DATASETS = ['abstracts', 'bmc', 'fly', 'hbc', 'news', 'wiki', 'merged_data']
URL = 'http://people.rc.rit.edu/~bsm9339/corpora/szeged_uncertainty/'


def download():
    print('Download Training Data Files')
    for (index, dataset) in enumerate(DATASETS):
        url = urljoin(URL, dataset)
        urlretrieve(url, dataset)
        print('  [{}/{}] {}'.format(index + 1, len(DATASETS), dataset))

if __name__ == "__main__":
    download()



import collections
import csv
import itertools
import operator
import os
import pickle
import re

GREEK_LOWER = re.compile(u'[αβγδεζηθικλμξπρσςτυφψω]')
GREEK_UPPER = re.compile(u'[ΓΔΘΛΞΠΣΦΨΩ]')
ENGLISH_LOWER = re.compile(r'[a-z]')
ENGLISH_UPPER = re.compile(r'[A-Z]')
DIGIT = re.compile(r'[0-9]')
ROMAN_LOWER = re.compile(
        'm{0,4}(cm|cd|d?c{0,3})(xc|xl|l?x{0,3})(ix|iv|v?i{0,3})'
    )
ROMAN_UPPER = re.compile(
        'M{0,4}(CM|CD|D?C{0,3})(XC|XL|L?X{0,3})(IX|IV|V?I{0,3})'
    )


def aggregate(predictions):
    prediction = 'C'

    frequency = collections.Counter(predictions)
    del frequency['C']

    if frequency:
        predictions = dict()
        for (p, f) in frequency.items():
            if f not in predictions:
                predictions[f] = list()
            predictions[f].append(p)
        predictions = sorted(
                predictions.items(), key=operator.itemgetter(0), reverse=True
            )
        prediction = predictions[0][1][0] \
            if predictions and len(predictions[0][1]) == 1 else 'U'

    return prediction


def dump(obj, filepath):
    with open(filepath, 'wb') as file:
        pickle.dump(obj, file)


def get_context(index, elements, size=2):
    if index < 0 or index >= len(elements):
        raise Exception('index must be positive and within elements\' bounds.')

    context = [None] * (size * 2)

    offsets = list(range(-size, size + 1))
    offsets.remove(0)

    indices = [index + offset for offset in offsets]
    for (i, j) in enumerate(indices):
        if j >= 0 and j < len(elements):
            context[i] = elements[j]

    return collections.OrderedDict(zip(offsets, context))


def get_wordpattern(word):
    pattern = ''.join(get_charpattern(character) for character in word)
    return ''.join(c for c, _ in itertools.groupby(pattern))


def get_charpattern(character):
    # TODO: How do we distinguish between a sequence of Roman Numerals and
    #       a sequence of letters?
    if not character.isalnum():
        return '!'
    elif ENGLISH_UPPER.search(character):
        return 'A'
    elif ENGLISH_LOWER.search(character):
        return 'a'
    elif DIGIT.search(character):
        return '0'
    elif GREEK_UPPER.search(character):
        return 'G'
    elif GREEK_LOWER.search(character):
        return 'g'
    elif ROMAN_UPPER.search(character):
        return 'R'
    elif ROMAN_LOWER.search(character):
        return 'r'


def get_features(data_file):
    """
    This preprocessing function takes in a data_file containing one sentence
    per line, generates the features for each token, and writes them to a file
    with the same name and '.tsv' appended to the end. See the documents
    /test_data.txt and /test_data.txt.tsv for examples of valid input and
    output, respectively.
    This preprocessing step allows for the features of a corpus to be generated
    only once, greatly speeding up the classification process.
    NOTE: The output file will be much larger in size than the input file.
    """
    with open(data_file, 'r') as f:
        filename = data_file + '.tsv'
        with open(filename, 'w', newline='') as tsvfile:
            tsv_writer = csv.writer(tsvfile, delimiter='\t', quotechar='|',
                                    quoting=csv.QUOTE_MINIMAL)
            for i, line in enumerate(f.readlines()):
                if line.strip() == '':
                    tsv_writer.writerow('')
                else:
                    features = get_features(line)
                    for key, val in sorted(features.items()):
                        (tok_num, tok) = key.split("_")
                        row = ['sent' + str(i) + "token" + str(tok_num),
                               str(tok), str(STEMMER.stem(tok)).lower(),
                               pos_tag(tok)[0][1], 'X', 'X']
                        for k, v in sorted(val.items()):
                            row.append(str(k) + ":" + str(v))
                        tsv_writer.writerow(row)


def get_verbs(filepath):
    """
    Return the contents of verbs file pointed to by the filepath argument as a
    dictionary in which the key is the conjugate of a verb and the value is
    uninflected verb form of the conjugate verb.

    For example, {'scolded': 'scold', 'scolding': 'scold'}

    Adapted from code provided in NodeBox:
    https://www.nodebox.net/code/index.php/Linguistics#verb_conjugation
    """
    verbs = dict()
    with open(filepath) as file:
        reader = csv.reader(file)
        for row in reader:
            for verb in row[1:]:
                verbs[verb] = row[0]

    return verbs


def load(filepath):
    if not os.path.exists(filepath):
        raise FileNotFoundError('No such file: {}'.format(filepath))

    with open(filepath, 'rb') as file:
        return pickle.load(file)




from .word import *

__all__ = ['Sentence', 'Sentences']


class Sentence(object):
    @classmethod
    def from_lines(cls, lines):
        instance = cls()

        instance.words = Words.from_lines(lines)

        instance.sentence = list()
        for word in instance.words.words:
            instance.sentence.append(word.word)

        return instance

    @classmethod
    def from_groups(cls, groups):
        instance = cls()

        instance.words = Words.from_groups(groups)

        instance.sentence = list()
        for word in instance.words.words:
            instance.sentence.append(word.word)

        return instance

    def get_words(self):
        return self.words.get_words()

    def get_sentence(self):
        return " ".join(self.sentence)

    def get_features(self):
        return self.words.get_features()

    def get_label(self, binary=True):
        labels = self.words.get_labels(binary=binary)
        if binary:
            for label in labels:
                if label == "X":
                    return "X"
                elif label != "C":
                    return "U"
            return "C"
        else:
            labs = {"C": 0, "U": 0, "I": 0, "N": 0, "E": 0, "D": 0}
            for label in labels:
                if label == "X":
                    return "X"
                else:
                    labs[label] += 1

            if (
                    labs["U"] != 0 or labs["E"] != 0 or labs["I"] != 0 or
                    labs["D"] != 0 or labs["N"] != 0
            ):

                labs.pop("C", None)
                max_val = max(labs.values())
                max_keys = []
                for k, v in labs.items():
                    if v == max_val:
                        max_keys.append(k)

                if len(max_keys) == 1:
                    return max_keys[0]
                else:
                    return "U"
            else:
                return "C"

    def get_data(self, binary=True):
        X, y, z = self.words.get_data(binary=True)
        return X, y, z


class Sentences(object):
    @classmethod
    def from_lineslist(cls, lineslist):
        instance = cls()

        instance.sentences = list()
        for lines in lineslist:
            instance.sentences.append(Sentence.from_lines(lines))

        return instance

    def get_data(self, binary=True):
        X, y = list(), list()
        for sentence in self.sentences:
            X.append(sentence)
            y.append(sentence.get_label(binary=binary))
        return X, y



from setuptools import setup

setup(
        name='uncertainty',
        version='0.1.0',
        packages=[
            'uncertainty', 'uncertainty.data', 'uncertainty.lib',
            'uncertainty.lib.nlp'
        ],
        package_data={
            'uncertainty': ['models/*.p', 'vectorizers/*.p'],
            'uncertainty.lib.nlp': ['verbs.txt']
        },
        install_requires=[
            'numpy==1.12.1',
            'scipy==0.19.0',
            'scikit-learn==0.18.1',
            'nltk==3.2.2'
        ],
        license='The MIT License (MIT) Copyright (c) 2017 Benjamin S. Meyers',
        description='Python implementation of a classifier for linguistic '
                    'uncertainty, based on the work by Vincze et al.',
        author='Benjamin S. Meyers',
        author_email='bsm9339@rit.edu',
        url='https://github.com/meyersbs/uncertainty',
        #download_url='https://github.com/meyersbs/uncertainty/uncertainty/dist/uncertainty-0.1.0.tar.gz',
        test_suite='uncertainty.tests',
        classifiers=[
            'Development Status :: 4 - Beta',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: MIT License',
            'Natural Language :: English',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Topic :: Scientific/Engineering :: Information Analysis',
            'Topic :: Software Development :: Libraries :: Python Modules',
            'Topic :: Text Processing :: Linguistic'
        ],
        keywords=[
            'nlp', 'natural language', 'natural language processing',
            'uncertainty', 'linguistic uncertainty', 'classification'
        ]
    )



import os

from datetime import datetime as dt

from nltk.stem import WordNetLemmatizer
from nltk.corpus import wordnet

from ... import helpers
from . import postagger, VERBS_PATH

VERBS = helpers.get_verbs(VERBS_PATH)
WORDNET_POS = {
    'N': wordnet.NOUN, 'V': wordnet.VERB, 'J': wordnet.ADJ, 'R': wordnet.ADV
}
MAP = {
    "'m": 'am', "'ll": 'will', "n't": 'not', "'ve": 'have', "'re": 'are'
}

lemmatizer = WordNetLemmatizer()


class Lemmatizer(object):
    """ Interface. """
    def __init__(self, tokens):
        """ Constructor. """
        self.tokens = tokens

    def execute(self):
        """ Raises NotImplementedError. """
        raise NotImplementedError("Lemmatizer is an abstract class. In must "
                                  "be implemented by another class. Try using "
                                  "the NLTKLemmatizer.")


def fix(token, lemma, prev=None, next=None):
    """
    Attempts to fix lemmatization errors with hardcoded rules.
    """
    if not token and not lemma and not prev and not next:
        raise ValueError("Recieved invalid input to lemmatizer.fix()")
    elif token.lower() == "ca":
        if next and next[0] and next[0].lower() == "n't":
            return "can"
        else:
            return lemma.lower()
    elif token.lower() == "as":
        return "as"
    elif token.lower() == "left":
        if prev and prev[1] == wordnet.VERB:
            return "leave"
        else:
            return lemma.lower()
    elif token in MAP:
        return MAP[token]
    elif lemma in VERBS:
        return VERBS[lemma]
    elif token in VERBS:
        return VERBS[token]
    else:
        return lemma.lower()


class NLTKLemmatizer(Lemmatizer):
    """ Implements Lemmatizer. """
    def __init__(self, tokens):
        """ Constructor. """
        super().__init__(tokens)

    def execute(self):
        """ Return a list of all tokens within the specified string. """
        lemmas = []

        tokens = [
                (t, WORDNET_POS.get(p[0], wordnet.NOUN))
                for (t, p) in postagger.PosTagger(self.tokens).execute()
            ]

        for (i, (token, pos)) in enumerate(tokens):
            lemma = lemmatizer.lemmatize(token, pos)
            prev = None if i == 0 else tokens[i - 1]
            next = None if i == len(tokens) - 1 else tokens[i + 1]
            lemmas.append(fix(token, lemma, prev, next))

        return lemmas



from collections import Counter
from operator import itemgetter
from os import path

from sklearn.feature_extraction import DictVectorizer
from sklearn.linear_model import LogisticRegression

from . import constants, helpers, word
from .lib.nlp import summarizer


class Classifier(object):
    def __init__(self, granularity='word', binary=False):
        self.granularity = granularity
        self.binary = binary

        self.classifier = None
        self.vectorizer = None

    def train(self, filepath):
        if not path.exists(filepath):
            raise FileNotFoundError('No such file: {}'.format(filepath))

        lines = None
        with open(filepath) as file:
            lines = file.readlines()

        words = word.Words.from_lines(lines)
        X, y, _ = words.get_data(binary=self.binary)
        self.vectorizer = DictVectorizer()
        X = self.vectorizer.fit_transform(X)
        if self.binary:
            self.classifier = LogisticRegression()
        else:
            self.classifier = LogisticRegression(
                    solver='newton-cg', multi_class='multinomial'
                )
        self.classifier.fit(X, y)
        self._dump()

    def predict(self, data):
        if self.classifier is None or self.vectorizer is None:
            self._load()

        groups = None
        if type(data) is str:
            groups = summarizer.Summarizer(data).execute()
        else:
            groups = data
        words = word.Words.from_groups(groups)
        X, _, _ = words.get_data(binary=self.binary)
        X = self.vectorizer.transform(X)
        y_pred = list(self.classifier.predict(X))

        return helpers.aggregate(y_pred) if self.granularity == 'sentence' \
            else y_pred

    # Properties

    @property
    def classifier_path(self):
        return constants.BCLASS_CLASSIFIER_PATH if self.binary \
               else constants.MCLASS_CLASSIFIER_PATH

    # Private Method

    def _dump(self):
        helpers.dump(self.classifier, self.classifier_path)
        helpers.dump(self.vectorizer, constants.VECTORIZER_PATH)

    def _load(self):
        self.classifier = helpers.load(self.classifier_path)
        self.vectorizer = helpers.load(constants.VECTORIZER_PATH)



import csv
import json
import traceback
import re
import sys
import warnings

from ..sentence import *
from ..word import *

warnings.filterwarnings("ignore", category=FutureWarning, module="__main__")

################################################################################
#### HELPER FUNCTIONS ##########################################################
################################################################################

def _clean_sent(sent):
    """
    This preprocessing function takes in a sentence object, grabs just the
    actual sentence, and runs three regex substitutions to clean up the
    sentence for querying.
    """
    s = re.sub(r'\s([\.,!?\)\]])', r'\1', sent.get_sentence().lower())
    s = re.sub(r'([\(\[])\s', r'\1', s)
    s = re.sub(r'(\s’[\ss])', r'’', s, re.UNICODE)
    return s

def _get_sent_label(labels, sent):
    """
    Given a list of uncertainty cues and a sentence object, return a label
    representing the classification of the entire sentence and each token within
    that sentence.

    Favor is given to the five uncertainty classes; if any token within the
    sentence is labeled as 'U', 'E', 'I', 'D', or 'N', the entire sentence is
    considered to be whichever label that occurs most frequently within the
    sentence. This decision is based on that made by Vincze et al. in their
    binary classifier. Only if there are no occurences of the five uncertainty
    labels within a sentence is the sentence classified as 'C'.
    """
    label_map = {"possible": "E", "probable": "E", "epistemic": "E",
                 "doxastic": "D", "investigation": "I", "condition": "N",
                 "certain": "C", "uncertain": "U"}
    word_labels = []
    for i, word in enumerate(sent.get_words()):
        word_labels.append(word.binary_label)
        if type(labels) != str:
            for k, v in labels.items():
                if word.word in v:
                    word_labels[i] = label_map[k.strip().rstrip("_").split("_")[-1].strip()]
                    break

    print("RAW:\t" + str(labels))
    if not bool(labels):
        return sent.get_label(), word_labels
    else:
        labs = {"E": 0.0, "D": 0.0, "I": 0.0, "N": 0.0, "C": 0.0, "U": 0.0}

        for k, v in labels.items():
            label = label_map[k.strip().rstrip("_").split("_")[-1].strip()]
            labs[label] += len(v)

        max_val = max(labs.values())
        max_keys = []
        for k, v in labs.items():
            if v == max_val:
                max_keys.append(k)

        if len(max_keys) == 1:
            return max_keys[0], word_labels
        else:
            return "U", word_labels

def _get_lines(filepath):
    """ Given a filepath, return a list of lines within that file. """
    lines = None
    with open(filepath) as file:
        lines = file.readlines()
    return lines


def _get_sentences(filepath):
    """ Given a filepath, return a list of sentences within that file. """
    sentences = list()
    _lines = list()
    for line in _get_lines(filepath):
        if line.strip() == '':  # End of Sentence
            sentences.append(_lines)
            _lines = list()
            continue
        _lines.append(line)
    return sentences

################################################################################
#### MAIN FUNCTION #############################################################
################################################################################

def merge_data(json_data, tagged_data):
    """
    There are two datasets associated with the work of Vincze et al [0].

    The first [1A] consists of a collection of XML objects that contain the text
    of a sentence and a list of uncertainty cues (with their categorization)
    that are within the sentence, if any. It has been parsed into a more
    intuitive JSON [1B] structure (at the expense of harddrive space).

    The second [2] consists of tab-delineated lists of tokens from [1A]. Each
    line in the files contains at least five columns: 1) an ID, 2) the raw
    token, 3) the lemma of the token, 4) the part-of-speech tag for the token,
    and 5) a label of certain or uncertain for the token. The tokens are in
    order by sentence, with each sentence separated by an empty line. All
    columns following the first five contain preparsed features used to train
    the binary classifier described in Vincze et al [0].

    This function parses both datasets (which contain the same sentences), and
    matches the uncertainty cues from [1B] to the tokens in [2]. This is not a
    trivial task; since [2] does not actually contain the raw sentences, the
    tokens had to be parsed into sentences and compared against those in [1B].

    This function results in a new file /data/merged_data [3] that is formatted
    in the same manner as [2], but contains a sixth column denoting a specific
    type of uncertainty (epistemic, doxastic, investigation, condition, other)
    when applicable. This new file [3] was used to train the multiclass models
    contained in this codebase.

    [0 ] http://doktori.bibl.u-szeged.hu/2291/1/Vincze_Veronika_tezis.pdf
    [1A] http://rgai.inf.u-szeged.hu/project/nlp/uncertainty/uncertainty.zip
    [1B] http://people.rc.rit.edu/~bsm9339/corpora/szeged_uncertainty/szeged_uncertainty_json.tar.gz
    [2 ] http://rgai.inf.u-szeged.hu/project/nlp/uncertainty/clexperiments.zip
    [3 ] http://people.rc.rit.edu/~bsm9339/corpora/szeged_uncertainty/merged_data
    """
    try:
        json_dict = json.loads(open(json_data, 'r').read())
        json_dict2 = {}
        cnt = 0
        for item in json_dict:
            cnt += 1
            if bool(item['ccue']):
                json_dict2.update({item['text'].lower(): item['ccue']})

        print("Found " + str(cnt) + " documents with uncertainty cues.")

        sents = Sentences.from_lineslist(_get_sentences(tagged_data))
        X, y = sents.get_data()

        with open(tagged_data + ".new", "w", newline='') as tsvfile:
            tsv_writer = csv.writer(tsvfile, delimiter='\t')
            for i, sent in enumerate(X):
                s = _clean_sent(sent)
                tags = {}
                if s in json_dict2.keys():
                    tags = json_dict2[s]
                elif re.sub(r"(\d\s)-(\s\d)", r"\1-\2", s) in json_dict2.keys():
                    tags = json_dict2[re.sub(r"(\d\s)-(\s\d)", r"\1-\2", s)]
                elif re.sub(r"\s/\s", r"/", s) in json_dict2.keys():
                    tags = json_dict2[re.sub(r"\s/\s", r"/", s)]
                elif re.sub(r'\\u00b1', '±', s) in json_dict2.keys():
                    tags = json_dict2[re.sub(r'\\u00b1', '±', s)]
                elif re.sub(r'\s-\s', '-', s) in json_dict2.keys():
                    tags = json_dict2[re.sub(r'\s-\s', '-', s)]
                elif re.sub(r"\\", '', s) in json_dict2.keys():
                    tags = json_dict2[re.sub(r"\\", '', s)]
                else:
                    tags = {}

                rows = []
                sent_label, word_labels = _get_sent_label(tags, sent)
                prepend = "000"
                for j, word in enumerate(sent.get_words()):
                    if j > 999:
                        prepend=""
                    elif j > 99:
                        prepend="0"
                    elif j > 9:
                        prepend="00"

                    row = ['sent' + str(i) + "token" + prepend + str(j),
                           str(word.word), str(word.root), str(word.pos)]

                    row.append(word.binary_label) # Binary Label
                    row.append(word_labels[j]) # Multiclass Label

                    for k, v in word.get_features().items():
                        row.append(str(k) + ":" + str(v))

                    tsv_writer.writerow(row)
                tsv_writer.writerow([""])

    except Exception as e:
        extype, exvalue, extrace = sys.exc_info()
        traceback.print_exception(extype, exvalue, extrace)
        return False

    return True
